###############################################################
# HTML FILTER ROUTINES
#
# Charlie KEMPSON - charlie@siren.demon.co.uk
# http://public.logica.com/~kempsonc/mh_index.htm
# Version 1.2
#
# This file contains the import filter information for parsing
# of SMGL compliant HTML files.  It provides the standard
# routines:
#
#        - MH_InterpretElement
#        - MH_InterpretEntity
#
# Plus a number of routines used by these standard routines
# to perform the formatting.
#
###############################################################

###############################################################
#
#    Copyright (c) 1996 Charlie Kempson
#
#    This program is free software; you can redistribute it 
#    and/or modify it under the terms of the GNU General 
#    Public License as published by the Free Software 
#    Foundation (version 2 of the License).
#
#    This program is distributed in the hope that it will 
#    be useful, but WITHOUT ANY WARRANTY; without even the 
#    implied warranty of MERCHANTABILITY or FITNESS FOR A 
#    PARTICULAR PURPOSE.  See the GNU General Public License 
#    for more details.
#
#    For a copy of the GNU General Public License, write to the 
#    Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
#    MA 02139, USA.
###############################################################

# SOME ASCII CODES
#
# 149 - A blob which could mark list items
# 153 - Trade Mark symbol
# 169 - Copyright symbol
# 151 - Line suitable for creating horizontal rules

# Create a horizontal rule
for {set i 0} {$i < 30} {incr i} {append gs_hrule [format "%c" 151]}

###############################################################
# The procedure for decoding tagged sequences
proc MH_InterpretElement { widget markup_code begin end text\
   search_start search_end {is_run_time 1} {image_path ""}} {

   # widget - the text widget to search
   # markup_code - the code found
   # begin - start of affected text
   # end - end of affected text
   # text - content of affected text
   # search_start search_end - search range
   global FONT_BOLD FONT_ITALIC FIXED_FONT gs_hrule
   set markup_deleted 0
   set markup_added 0

   # NOTE: Input code is in upper case
   switch $markup_code {

      ######################################
      # Allowed markup codes
      ######################################

      H1 -
      H2 -
      H3 -
      TITLE  {
         # Use generic routine
         incr markup_deleted [MH_FormatChars $widget $markup_code \
            "-font $FONT_BOLD -underline True -justify center" \
            $begin $end $text $search_start $search_end]
      }
      B  {
         # Use generic routine
         incr markup_deleted [MH_FormatChars $widget $markup_code \
            "-font $FONT_BOLD" \
            $begin $end $text $search_start $search_end]
      }
      I  {
         # Use generic routine
         incr markup_deleted [MH_FormatChars $widget $markup_code \
            "-font $FONT_ITALIC" \
            $begin $end $text $search_start $search_end]
      }
      TT  {
         # Use generic routine
         incr markup_deleted [MH_FormatChars $widget $markup_code \
            "-font $FIXED_FONT" \
            $begin $end $text $search_start $search_end]
      }
      CENTER  {
         # Use generic routine
         incr markup_deleted [MH_FormatChars $widget $markup_code \
            "-justify center" \
            $begin $end $text $search_start $search_end]
      }
      IMG  {
         incr markup_deleted [MH_FormatImage $widget \
            $begin $end $text $image_path]
         incr markup_added
      }
      XREF  {
         incr markup_deleted [MH_FormatXref $widget $markup_code \
            $begin $end $text $search_start $search_end $is_run_time]
      }
      P  {
         incr markup_deleted [MH_FormatPara $widget $markup_code \
            "\n\n" $begin $end $text $search_start $search_end]
         incr markup_added 2
      }
      UL -
      DL -
      BR  {
         # Use generic routine
         incr markup_deleted [MH_FormatReplace $widget $markup_code \
            "\n" $begin $end $text $search_start $search_end]
         incr markup_added 1
      }
      HR  {
         # Use generic routine
         incr markup_deleted [MH_FormatReplace $widget $markup_code \
            "\n\n<CENTER>${gs_hrule}</CENTER>\n\n" $begin $end $text $search_start $search_end]
         incr markup_added 0
      }
      DT - 
      LI  {
         # Use generic routine
         set litem [format "%c" 149]
         incr markup_deleted [MH_FormatReplace $widget $markup_code \
            "\n${litem}   " $begin $end $text $search_start $search_end]
         incr markup_added 5
      }
      DD  {
         # Use generic routine
         incr markup_deleted [MH_FormatReplace $widget $markup_code \
            "\n\t" $begin $end $text $search_start $search_end]
         incr markup_added 2
      }
      HTML -
      HEAD -
      BODY -
      A  - 
      LINK -
      ADDRESS -
      TABLE -
      TR -
      TD -
      FONT  {
         # Use generic routine
         incr markup_deleted [MH_FormatReplace $widget $markup_code \
            "" $begin $end $text $search_start $search_end]
         incr markup_added 0
      }
      default {
         #Info_Dialog . "Format '$markup_code' is not valid - skipping"
         return ""
      }
   }

   # Return the number of characters deleted
   return [list $markup_added $markup_deleted]
}

###############################################################
# The procedure for decoding entities (special characters or 
# sequences)
proc MH_InterpretEntity { widget begin end text \
   {is_run_time 1} {image_path ""}} {

   # NOTE: Input text has been converted to upper case
   set markup_deleted 0
   set markup_added 0

   #Info_Dialog . $text
   switch $text {

      ############################
      # Allowed special characters
      ############################
      AMPER     {set special_char \& }
      COPY      {set special_char [format "%c" 169]}
      default   {return ""}
   }
   if {$special_char != ""} {
      # Replace special character code with the character
      $widget delete $begin "$end +1c"
      $widget insert $begin $special_char

      # And modify character count
      incr markup_deleted [string length $text]
      incr markup_added [string length $special_char]
   }

   # Return the number of characters deleted
   return [list $markup_added $markup_deleted]
}
