###############################################################
# MINIHELP FILTER ROUTINES
#
# Charlie KEMPSON - charlie@siren.demon.co.uk
# http://public.logica.com/~kempsonc/mh_index.htm
# Version 1.2
#
# This file contains the import filter information for parsing
# of SMGL compliant MiniHelp files.  It provides the standard
# routines:
#
#        - MH_InterpretDocument
#        - MH_InterpretElement
#        - MH_InterpretEntity
#
# Plus a number of routines used by these standard routines
# to perform the formatting.
#
###############################################################

###############################################################
#
#    Copyright (c) 1996 Charlie Kempson
#
#    This program is free software; you can redistribute it 
#    and/or modify it under the terms of the GNU General 
#    Public License as published by the Free Software 
#    Foundation (version 2 of the License).
#
#    This program is distributed in the hope that it will 
#    be useful, but WITHOUT ANY WARRANTY; without even the 
#    implied warranty of MERCHANTABILITY or FITNESS FOR A 
#    PARTICULAR PURPOSE.  See the GNU General Public License 
#    for more details.
#
#    For a copy of the GNU General Public License, write to the 
#    Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
#    MA 02139, USA.
###############################################################

# SOME ASCII CODES
#
# 149 - A blob which could mark list items
# 153 - Trade Mark symbol
# 169 - Copyright symbol
# 151 - Line suitable for creating horizontal rules

# Some useful globals
set g_lost_page_num 1

###############################################################
# Procedure called during load of the file.  Interprets
# tagged data found and creates pages from the input data.
# This function must return the start of the next search.
proc MH_InterpretDocument {file widget code begin end text} {

   # file - the global variable containing the file data
   # widget - the text widget to search
   # code - the markup code found
   # begin - start of affected text
   # end - end of affected text
   # text - content of affected text
   # Info_Dialog . $file 

   # NOTE: Input code is in upper case
   switch $code {

      ######################################
      # Allowed markup codes
      ######################################
      MINIHELP11  {
         MH_ExtractFile $file $widget $code $begin $end $text
         return $begin
      }
      PAGE  {
         MH_ExtractPage $file $widget $code $begin $end $text
         return $begin
      }
      default {
         return "$begin +1c"
      }
   }
}

###############################################################
# The procedure for decoding tagged sequences
proc MH_InterpretElement { widget markup_code begin end text\
   search_start search_end {is_run_time 1} {image_path ""}} {

   # widget - the text widget to search
   # markup_code - the code found
   # begin - start of affected text
   # end - end of affected text
   # text - content of affected text
   # search_start search_end - search range
   global FONT_BOLD FONT_ITALIC FIXED_FONT
   set markup_deleted 0
   set markup_added 0

   # NOTE: Input code is in upper case
   switch $markup_code {

      ######################################
      # Allowed markup codes
      ######################################

      TITLE  {
         # Use generic routine
         incr markup_deleted [MH_FormatChars $widget "TITLE" \
            "-font $FONT_BOLD -underline True -justify center" \
            $begin $end $text $search_start $search_end]
      }
      B  {
         # Use generic routine
         incr markup_deleted [MH_FormatChars $widget "B" \
            "-font $FONT_BOLD" \
            $begin $end $text $search_start $search_end]
      }
      I  {
         # Use generic routine
         incr markup_deleted [MH_FormatChars $widget "I" \
            "-font $FONT_ITALIC" \
            $begin $end $text $search_start $search_end]
      }
      TT  {
         # Use generic routine
         incr markup_deleted [MH_FormatChars $widget "TT" \
            "-font $FIXED_FONT" \
            $begin $end $text $search_start $search_end]
      }
      UL  {
         # Use generic routine
         incr markup_deleted [MH_FormatChars $widget "UL" \
            "-underline True" \
            $begin $end $text $search_start $search_end]
      }
      CENTER  {
         # Use generic routine
         incr markup_deleted [MH_FormatChars $widget "CENTER" \
            "-justify center" \
            $begin $end $text $search_start $search_end]
      }
      IMG  {
         incr markup_deleted [MH_FormatImage $widget \
            $begin $end $text $image_path]
         incr markup_added
      }
      XREF  {
         incr markup_deleted [MH_FormatXref $widget "XREF" \
            $begin $end $text $search_start $search_end $is_run_time]
      }
      P  {
         incr markup_deleted [MH_FormatPara $widget "P" \
            "\n\n" $begin $end $text $search_start $search_end]
         incr markup_added 2
      }
      BR  {
         # Use generic routine
         incr markup_deleted [MH_FormatReplace $widget "BR" \
            "\n" $begin $end $text $search_start $search_end]
         incr markup_added 1
      }
      PAGE  {
         # Use generic routine
         incr markup_deleted [MH_FormatReplace $widget "PAGE" \
            "" $begin $end $text $search_start $search_end]
      }

      default {
         #Info_Dialog . "Format '$markup_code' is not valid - skipping"
         return ""
      }
   }

   # Return the number of characters deleted
   return [list $markup_added $markup_deleted]
}

###############################################################
# The procedure for decoding entities (special characters or 
# sequences)
proc MH_InterpretEntity { widget begin end text \
   {is_run_time 1} {image_path ""}} {

   # NOTE: Input text has been converted to upper case
   set markup_deleted 0
   set markup_added 0

   #Info_Dialog . $text
   switch $text {

      ############################
      # Allowed special characters
      ############################
      DOLLAR    {set special_char \$ }
      AMPER     {set special_char \& }
      LBRACE    {set special_char \{ }
      RBRACE    {set special_char \} }
      LSQUARE   {set special_char \[ }
      RSQUARE   {set special_char \] }
      QUOT      {set special_char \" }
      SQUOT     {set special_char \' }
      BACKSLASH {set special_char \\ }
      default   {return ""}
   }
   if {$special_char != ""} {
      # Replace special character code with the character
      $widget delete $begin "$end +1c"
      $widget insert $begin $special_char

      # And modify character count
      incr markup_deleted [string length $text]
      incr markup_added [string length $special_char]
   }

   # Return the number of characters deleted
   return [list $markup_added $markup_deleted]
}

###############################################################
# The procedure for formatting characters
proc MH_FormatXref { widget markup_key begin end text \
   search_start search_length {is_run_time 1} } {
   # Info_Dialog . "Key: $markup_key, Tag1: $begin ->$end, Text: $text"

   # widget - text widget containing the area to be formatted
   # markup_key - the tag
   # begin - the start location
   # end - the end location (actually formats up to one char before)
   # text - the text within the first tag (contains the link name)
   # search_start - search_length - search parameters
   # is_run_time - boolean (1 or 0) determining bindings
   global GREEN RED FONT_BOLD
   set markup_deleted 0

   # Find the cross reference page
   set attribute_list [MH_GetAttributes $text]
   if {[llength $attribute_list] == 2} {
      switch [lindex $attribute_list 0] {
         PAGE {
            set link_name [lindex $attribute_list 1]
            if {$link_name == ""} {return}
         }
      }
   }  else {
      # No valid link found - return
      return $markup_deleted
   }

   set mark [MH_FindEndTag $widget $markup_key \
      $end "$search_start +${search_length}c"]
   set mark2(begin) [lindex $mark 0]
   if {$mark2(begin) != ""} {
      # End tag has been found
      set mark2(end) [lindex $mark 1]
      set mark2(text) [lindex $mark 2]
      $widget tag add $link_name $begin "$mark2(end) +1c"
      $widget tag raise $link_name
      $widget tag configure $link_name -foreground \
         $GREEN -font $FONT_BOLD
      $widget tag bind $link_name <Enter> \
         "$widget tag configure $link_name -foreground $RED"
      $widget tag bind $link_name <Leave> \
         "+$widget tag configure $link_name -foreground $GREEN"
      if {$is_run_time == 1} {
         $widget tag bind $link_name <ButtonPress> "MH_MiniHelp $link_name"
      } else {
         $widget tag bind $link_name <ButtonPress> "MH_ShowHyperMenu $link_name"
      }

      # Perform tag deletion and bookkeeping
      $widget delete $mark2(begin) "$mark2(end) +1c"
      $widget delete $begin "$end +1c"
      incr markup_deleted [expr [string length $text] \
         + [string length $mark2(text)] +2]
   }

   # Return new markup deleted count
   return $markup_deleted
}

###############################################################
# Procedure to extract a file from the SGML file
# Returns 0 if sucessful and -1 otherwise
proc MH_ExtractFile {file widget code begin end text} {

   # Globals
   global gs_image_base_dir 

   # file - the global variable containing the file data
   # widget - the text widget to search
   # code - the markup code found
   # begin - start of affected text
   # end - end of affected text
   # text - content of affected text

   # Find any attributes of the page
   set attribute_list [MH_GetAttributes $text]
   set attribute_count 0
   set title "No Title"
   set imgpath ""
   while {$attribute_count < [llength $attribute_list]} {
      switch [lindex $attribute_list $attribute_count] {
         TITLE {
            set title [lindex $attribute_list \
               [expr $attribute_count +1]]
            #Info_Dialog . "Found title $title"
         }
         IMGPATH {
            set imgpath [lindex $attribute_list \
               [expr $attribute_count +1]]
            #Info_Dialog . "Found title $title"
         }
      }
      incr attribute_count 2
   }

   # Find the end of the page
   set mark [MH_FindEndTag $widget $code $end end]
   set mark2(begin) [lindex $mark 0]
   if {$mark2(begin) != ""} {
      # End tag has been found
      set mark2(end) [lindex $mark 1]
      set mark2(text) [lindex $mark 2]
      $widget delete $mark2(begin) "$mark2(end) + 1c"
   }

   # Set the page name and level into the structure.  Note that
   # the text is preserved including the page element.
   global $file
   set ${file}(Title) $title
   set ${file}(Path) $imgpath
   set gs_image_base_dir $imgpath
   $widget delete $begin "$end +1c"

   # Finished Ok
   return $begin
}

###############################################################
# Procedure to extract a page from the SGML file
# Returns 0 if sucessful and -1 otherwise
proc MH_ExtractPage {file widget code begin end text} {

   # file - the global variable containing the file data
   # widget - the text widget to search
   # code - the markup code found
   # begin - start of affected text
   # end - end of affected text
   # text - content of affected text

   # Setup basic information
   # BASE ELEMENT: gsSGML_<File> containing:
   # SUB ELEMENT <FullFileName>
   # SUB ELEMENT <DTD>
   # SUB ELEMENT <Path>
   # SUB ELEMENT <TOC>
   # SUB ELEMENT <PageName> containing:
   # SUB ELEMENT <Level>
   # SUB ELEMENT <Title>
   # SUB ELEMENT <Text>

   # Flags
   set found_page_name 0
   global g_lost_page_num $file
   set page_name "LostPage$g_lost_page_num"
   incr g_lost_page_num
   set lvl 1

   # Find any attributes of the page
   set attribute_list [MH_GetAttributes $text]
   set attribute_count 0
   while {$attribute_count < [llength $attribute_list]} {
      switch [lindex $attribute_list $attribute_count] {
         LVL {
            if [catch {scan [lindex $attribute_list \
               [expr $attribute_count +1]] "%d" lvl}] {
               # Failed to get number - default
               set lvl 1
            }
         }
         ID {
            set page_name [lindex $attribute_list \
               [expr $attribute_count +1]]
            #Info_Dialog . "Found page: $page_name"
            set found_page_name 1
         }
      }
      incr attribute_count 2
   }

   # Check that an ID was supplied...
   if {$found_page_name == 0} {
      Info_Dialog . "Found page with invalid/no ID, adding anyway..."
   }

   # Find the end of the page
   set mark [MH_FindEndTag $widget $code $end end]
   set mark2(begin) [lindex $mark 0]
   if {$mark2(begin) != ""} {
      # End tag has been found
      set mark2(end) [lindex $mark 1]
      set mark2(text) [lindex $mark 2]
   } else {
      # No end of page found! Assume end-of-file
      set mark2(begin) end
      set mark2(end) end
      set mark2(text) ""
   }

   # Set the page name and level into the structure.  Note that
   # the text is preserved including the page element.
   set ${file}(${page_name}_Level) $lvl
   set ${file}(${page_name}_Text) [$widget get "$begin" "$mark2(end) +1c"]
   $widget delete $begin "$mark2(end) +1c"
   lappend ${file}(TOC) $page_name

   # Finished Ok
   return $begin
}



