###############################################################
# MINIHELP HELP PARSER ROUTINES
# Charlie KEMPSON - charlie@siren.demon.co.uk
# http://public.logica.com/~kempsonc/mh_index.htm
# Version 1.2
#
# The following routines are provided for external functions
# to call:
#
# MH_ReadSGMLFile
# -------------------
# This routine is called with a filename.  The file is
# read and converted to the internal document format
# using the function MH_MarkupToInternal.  This function
# uses an external function MH_InterpretDocument to 
# paginate the text.
#
# MH_MarkupToText
# ---------------
# This takes as input a string containing the SGML formatted 
# data for a page, and places the formatted output in a specified 
# location within a multiline text widget, replacing the 
# data that was previously in the location.  If no replacement
# positions are specified, the entire text widget data is 
# replaced.  This function uses two external functions
# MH_InterpretElement and MH_InterpretEntity to parse
# the SMGL.
#
# MH_FindEndTag and MH_FindNextStartTag
# -------------------------------------
# This takes as input parameters a tag (e.g. TT) and search
# start and end locations and reports the location of the 
# found end tag.  It outputs a list "begin end text" 
# corresponding to the location and tagged text found.
# Note that the FindNextStartTag will fail if there is space
# between the start < and the tag.
#
# MH_GetAttributes
# ----------------
# Takes a tag with or without attributes, and returns
# a list of attributes along with their argments.  So
# IMG SRC="help.gif" ALT="Help!" would return a list 
# with the elements SRC "help.gif" ALT and "Help!"
# Note that the attributes must be separated from their
# arguments by = signs.
# 
# MH_FormatChars
# --------------
# This routine provides basic text formatting capabilities.
# You specify the tag and formatting, e.g. B and '-font
# $FONT_NORMAL -underline True -justify center' and the
# routine formats the specified text (having found the end
# tag) and returns the number of chars deleted.
#
# MH_FormatReplace
# ----------------
# This routine provides a general mechanism for dealing
# with elements which require the insertion of characters
# to replace formatting tags.  Example: break tags <BR>
# require insertion of a newline character.
#
# INTERFACE REQUIREMENTS
# ======================
#
# Note that MH_MarkupToText requires two additional functions
# which perform tagging according to specified strings.  This
# method allows different tagging systems, e.g. HTML, MiniHelp
# to be used.  The routines should be placed in import files
# e.g. minihelp_filter.tcl, which may be sourced as needed
# by the help parser.  Both functions should return a list
# containing the elements "added deleted", or "" if no action
# was taken.
#
# MH_InterpretElement - interprets <TAG>Text</TAG> sequences
# -------------------
# widget - The text widget to use
# markup_code - The markup code found
# mark1(begin) - The begin location for the tagged text
# mark1(end) - The begin location for the tagged text
# mark1(text) - The text, e.g. IMC SRC="help.gif"
# search_start - Search parameters
# search_length - Search parameters
# is_run_time - Run time flag - usually 1
# image_path - May in fact just be a path
#
# MH_InterpretEntity - interprets &dollar; sequences
# -------------------
# widget - The text widget to use
# mark1(begin) - The begin location for the tagged text
# mark1(end) - The begin location for the tagged text
# mark1(text) - The text, e.g. IMC SRC="help.gif"
# is_run_time - Run time flag - usually 1
# image_path - May in fact just be a path
#
###############################################################

###############################################################
#
#    Copyright (c) 1996 Charlie Kempson
#
#    This program is free software; you can redistribute it 
#    and/or modify it under the terms of the GNU General 
#    Public License as published by the Free Software 
#    Foundation (version 2 of the License).
#
#    This program is distributed in the hope that it will 
#    be useful, but WITHOUT ANY WARRANTY; without even the 
#    implied warranty of MERCHANTABILITY or FITNESS FOR A 
#    PARTICULAR PURPOSE.  See the GNU General Public License 
#    for more details.
#
#    For a copy of the GNU General Public License, write to the 
#    Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
#    MA 02139, USA.
###############################################################

# Globals for this module
set gs_sgml_structure ""

###############################################################
# Procedure to open an SGML file and convert it into pages
# which may be displayed.  Returns 0 if the file was read 
# OK, and -1 otherwise.
proc MH_ReadSGMLFile { file widget } {

   # The formats currently uderstood by this parser are:

   # MINIHELP SGML
   # -------------
   # Comprises a number of pages, each containing tagged
   # text.  The file is read into memory, then split into
   # pages.  The memory structure uses a single variable
   # with multiple elements:
   #
   # BASE ELEMENT: gsSGML_<File> containing:
   # SUB ELEMENT <FullFileName>
   # SUB ELEMENT <Path>
   # SUB ELEMENT <PageName> containing:
   # SUB ELEMENT <Level>
   # SUB ELEMENT <Title>
   # SUB ELEMENT <Text>

   # Get the file name
   if [file exists $file] {

      # Get the filename
      regsub {[`"#;:',.?/\|!$%^&*()|~<>,-]+} [file tail [file \
         rootname $file]] {} filename

      # Check whether we haven't already loaded this file
      if {[info globals "gsSGML_$filename"] != ""} {
         if [Question_Dialog . "A file with name $filename has already 
been loaded.  To you wish to overwrite it?" No Yes] {
            if [Question_Dialog . "Are you sure?" No Yes] {
               # Delete current information on the file
               global gsSGML_$filename
               unset gsSGML_$filename
            } else {return -1}
         } else {return -1}
      }
   } else {
      # File does not exist
      Info_Dialog . " File $file does not exist"
      return -1
   }

   # Read the file into memory for parsing.
   if [catch {set fd [open $file]} error] {
      Info_Dialog . "Unable to open $filename : $error"
      return -1}
   if [catch {set file_contents [string trim [read $fd]]} error] {
      Info_Dialog . "Unable to read $filename : $error"
      return -1}
   close $fd

   # Check the file type - the first part of the file
   # contains the DTD
   set doctype ""
   if ![scan $file_contents "<!DOCTYPE %s" doctype] {
      if ![scan $file_contents "<!doctype %s" doctype] {
         # Ok, file is not SGML compliant.  Check for
         # the first tag in the file.
         if {[scan $file_contents "<%s>" doctype] >= 1} {
            # Found something!
            set doctype [string trim $doctype "<>"]
         }
      }
   }
   switch [string toupper $doctype] {

      # FILE TYPE MINIHELP VERSION 1.1
      # ------------------------------
      MINIHELP11 {
         if [catch {source minihelp_filter.tcl}] {
            set dir [file dirname [info script]]
            source $dir/minihelp_filter.tcl
         }
         return [MH_MarkupToInternal $widget $file $file_contents]
      }

      # FILE TYPE HTML
      # --------------
      HTML {
         if [catch {source html_filter.tcl}] {
            set dir [file dirname [info script]]
            source $dir/html_filter.tcl
         }
         MH_MarkupToText $widget INSERT 0 [file dirname $file] \
            $file_contents
         if [catch {source minihelp_filter.tcl}] {
            set dir [file dirname [info script]]
            source $dir/minihelp_filter.tcl
         }
         return 0
      }

      default {
         Info_Dialog . "Document format $doctype not supported.
File load aborted."
         return -1
      }
   }
}

###############################################################
# The procedure for creating formatted help text from a page.
proc MH_MarkupToInternal { widget file {utext ""} } {

   # widget - the text widget to display results in
   # file - the filename of the text source
   # utext - the text to be processed

   # After this routine returns the current project pages
   # may be refered to using the global variable gs_sgml_structure

   # This is now a one stage process.  Delimiters as specified 
   # by the SGML guidelines, e.g. < and & are searched for
   # and are responded to appropriately.

   # The first thing to do is to insert the text into 
   # the widget.  Then the widget is withdrawn so that 
   # redraw is quicker.  All newlines and multiple
   # spaces are removed from the text - newlines are
   # input for paragraph markings.
   #set orig_state [$widget cget -state]
   #set orig_pack [pack info $widget]
   #pack forget $widget
   #set dialog [Working_Dialog . "Reading help file. Please wait..."]
   $widget configure -state normal
   $widget delete 1.0 end
   $widget insert 1.0 $utext

   # Setup basic information
   # BASE ELEMENT: gsSGML_<File> containing:
   # SUB ELEMENT <FullFileName>
   # SUB ELEMENT <Title>
   # SUB ELEMENT <Path>
   # SUB ELEMENT <TOC>
   # SUB ELEMENT <PageName> containing:
   # SUB ELEMENT <Level>
   # SUB ELEMENT <Title>
   # SUB ELEMENT <Text>
   set filename [file tail [file rootname $file]]
   regsub {[`"#;:',.?/\|!$%^&*()|~<>,-]+} [file tail [file \
      rootname $file]] {} filename
   global gsSGML_$filename gs_sgml_structure
   set gs_sgml_structure gsSGML_${filename}
   set gsSGML_${filename}(Title) "No Title"
   set gsSGML_${filename}(Path) ""
   set gsSGML_${filename}(FullFileName) $file
   set gsSGML_${filename}(LastPage) ""
   set gsSGML_${filename}(TOC) ""
   set gsSGML_${filename}(Modified) 0

   # Strip the header from the front of the file
   set mark1(begin) [$widget search "\]>" 1.0 end]
   if {$mark1(begin) == ""} {
      #Info_Dialog . "No DTD found... will try to continue"
      set gsSGML_${filename}(DTD) ""
   } else {
      set gsSGML_${filename}(DTD) [$widget get 1.0 \
         "$mark1(begin) +2c"]
      $widget delete 1.0 "$mark1(begin) +2c"
   }

   # Now search for the beginning of the keyword constructs.
   set search_start 1.0
   set mark1(begin) [$widget search "<" $search_start end]
   set loopcount 0
   while {$mark1(begin) != ""} {
      incr loopcount

      # Potential Markup - determine which...
      set mark1(end) [$widget search ">" $mark1(begin) end]
      if {$mark1(end) != ""} {
         # Looks good - Check the tag code
         set mark1(text) [$widget get "$mark1(begin) +1c" $mark1(end)]
         if {[scan $mark1(text) "%s" markup_code] == 1} {
            set search_start [MH_InterpretDocument gsSGML_${filename} \
               $widget [string toupper $markup_code] $mark1(begin) \
               $mark1(end) $mark1(text)]
         }
      } else {
         set search_start "$mark1(begin) +1c"
      }

      # And search for the next piece of markup
      set mark1(begin) [$widget search "<" $search_start end]
   }

   # Info_Dialog . "Parsing completed with $loopcount stops."
   return 0
}

###############################################################
# The procedure for creating formatted help text from a page.
proc MH_MarkupToText { widget type is_run_time image_path \
   {utext ""} {ibegin "1.0"} {iend "end"} } {

   # widget - the text widget to display results in
   # type - INSERT or REFORMAT.  If insert is specified the
   #    text is inserted into the text widget and formatted.
   #    otherwise the text in the widget is reformatted (i.e.
   #    utext should be "")
   # is_run_time - flag (1 or 0) indicating how the hyperlinks
   #    are displayed and bound to actions
   # image_path - the (optional) image path (list)
   # utext - the text to be processed (may be null)
   # ibegin - iend - the area of the text to be replaced

   # This is now a one stage process.  Delimiters as specified 
   # by the SGML guidelines, e.g. < and & are searched for
   # and are responded to appropriately.

   # The first thing to do is to insert the text into 
   # the widget.  Then the widget is withdrawn so that 
   # redraw is quicker.  All newlines and multiple
   # spaces are removed from the text - newlines are
   # input for paragraph markings.
   set orig_state [$widget cget -state]
   set orig_pack [pack info $widget]
   pack forget $widget
   if ![string compare $type "INSERT"] {
      regsub -all {  } $utext { } utext
      regsub -all \n $utext {} utext
      $widget configure -state normal
      $widget delete $ibegin $iend
      $widget insert $ibegin $utext
   } else {
      set utext [$widget get $ibegin $iend]
   }

   # Now begin at the insert location, and search for
   # the beginning of the keyword constructs.
   set search_start $ibegin
   set search_length [string length $utext]
   set mark1(begin) [$widget search -regexp \[<|\&\] $search_start \
      "$search_start +${search_length}c"]
   # Info_Dialog . "Start: $search_start, length: $search_start +${search_length}c"
   # Loop until no more markup is found
   set loopcount 0
   while {$mark1(begin) != ""} {

      # Keep track of added and deleted characters
      set markup_deleted 0
      set chars_added    0
      incr loopcount

      # Get the character at the mark
      switch [$widget get $mark1(begin)] {
         \& { 
            # Potential special character
            # End delimiter is a ; so search for that
            set mark1(end) [$widget search \; $mark1(begin) \
               "$search_start +${search_length}c"]
            if {$mark1(end) != ""} {
               # Looks good - Check the character code
               set mark1(text) [string toupper [$widget get \
                  "$mark1(begin) +1c" $mark1(end)]]
               set return [MH_InterpretEntity $widget \
                  $mark1(begin) $mark1(end) $mark1(text) \
                  $is_run_time $image_path]
               if {$return != ""} {
                  incr chars_added [lindex $return 0]
                  incr markup_deleted [lindex $return 1]
               }
            }
         }
         <  { 
            # Potential Markup - determine which...
            # End delimiter is a > so search for that
            set mark1(end) [$widget search ">" $mark1(begin) \
               "$search_start +${search_length}c"]
            if {$mark1(end) != ""} {
               # Looks good - Check the tag code
               set mark1(text) [$widget get "$mark1(begin) +1c" $mark1(end)]
               if {[scan $mark1(text) "%s" markup_code] == 1} {
                  set return [MH_InterpretElement $widget \
                     [string toupper $markup_code] \
                     $mark1(begin) $mark1(end) $mark1(text) \
                     $search_start $search_length $is_run_time \
                     $image_path]
                  if {$return != ""} {
                     incr chars_added [lindex $return 0]
                     incr markup_deleted [lindex $return 1]
                  }
               }
            }
         }
      }

      # And adjust the search length deleted_length
      set delete_from_start [string length [$widget get $search_start \
         "$mark1(begin) + 1c"]]
      if {$markup_deleted > 0} {
         set search_start $mark1(begin)
      } else {
         set search_start "$mark1(begin) +1c"
      }
      set search_length [expr $search_length - $delete_from_start \
         - $markup_deleted + $chars_added]
      set text [$widget get $search_start "$search_start +${search_length}c"]
      set mark1(begin) [$widget search -regexp \[<|\&\] $search_start \
         "$search_start +${search_length}c"]
#      Info_Dialog . "Start: $search_start, length: $search_start +${search_length}c
#Found: $mark1(begin)
#+++++++++++++++++++++++++ TEXT ++++++++++++++++++++++++++++++
#$text"
   }

   # And repack the text
   eval pack $widget $orig_pack
   eval $widget configure -state $orig_state
   # Info_Dialog . "Parsing completed with $loopcount stops."
   return 0
}

###############################################################
# The procedure for finding the end tag in a sequence
proc MH_FindEndTag { widget tag {begin "1.0"} {end "end"} } {

   # widget - the text widget to search
   # tag - the end tag text to find
   # begin - start of text to search
   # end - end of text to search

   # Note that this routine returns a null string in mark(begin)
   # if the end tag is not found - this must be checked in the 
   # calling routine.

   # Find the positions
   set mark(begin) [$widget search -nocase "</$tag>" $begin $end]
   if {$mark(begin) != ""} {
      set mark(end) [$widget search ">" $mark(begin) $end]
      set mark(text) "/$tag"

      # Return the results
      return [list $mark(begin) $mark(end) $mark(text)] 
   } else {
      # Return the results
      return ""
   }   
}

###############################################################
# The procedure for finding the end tag in a sequence
proc MH_FindNextStartTag { widget tag {begin "1.0"} {end "end"} } {

   # widget - the text widget to search
   # tag - the start tag text to find
   # begin - start of text to search
   # end - end of text to search

   # Note that this routine returns a null string in mark(begin)
   # if the end tag is not found - this must be checked in the 
   # calling routine.

   # Find the positions
   set mark(begin) [$widget search -nocase "<$tag" $begin $end]
   if {$mark(begin) != ""} {
      set mark(end) [$widget search ">" $mark(begin) $end]
      set mark(text) [$widget get $mark(begin) "$mark(end) +1c"]
      # Return the results
      return [list $mark(begin) $mark(end) $mark(text)] 
   } else {
      # Return the results
      return ""
   }   
}

###############################################################
# The procedure for getting the attributes out of the begin tag
proc MH_GetAttributes { tag } {

   # Takes a tag with or without attributes, and returns
   # a list of attributes along with their argments.  So
   # IMG SRC="help.gif" ALT="Help!" would return a list 
   # with the elements SRC "help.gif" ALT and "Help!"

   # Note that the attributes must be separated from their
   # arguments by = signs, and that the list returned has
   # the attributes in upper case, but the arguments in
   # the original case.
   set attribute_list ""

   set tag [string trim $tag]
   set pos [string first " " $tag]
   if {$pos > 0} {
      # Found a space, so examine the rest of the string
      set tag [string trim [string range $tag [expr $pos + 1] end]]
      set pos [string first "=" $tag]
      while {$pos > 0} {
         #Info_Dialog . $tag
         # Analyse the tag
         lappend attribute_list [string toupper [string trim \
            [string range $tag 0 [expr $pos -1]]]]
         set tag [string trim [string range $tag [expr $pos +1] end]]
         # Find the next whitespace, or the enclosing quotes
         if {[string index $tag 0] == "\""} {
            set tag [string range $tag 1 end]
            set end [string first "\"" $tag]
         } else {
            set end [string first " " $tag]
         }
         if {$end > 0} {
            lappend attribute_list [string trim \
               [string range $tag 0 $end] "\"\ \n\t"]
            set tag [string trim [string range $tag [expr $end +1] end]]
         } else {
            lappend attribute_list [string trim \
               [string range $tag 0 end] "\"\ \n\t"]
            set tag ""
         }
         # Find the next tag
         set pos [string first "=" $tag]
      }
   }

   # Return the list (or null string)
   # Info_Dialog . $attribute_list
   return $attribute_list
}

###############################################################
# The procedure for formatting characters
proc MH_FormatChars { widget markup_key markup_format begin end \
   text search_start search_length} {
   #Info_Dialog . "Key: $markup_key, Tag1: $begin ->$end"

   # widget - text widget containing the area to be formatted
   # tag - the string to tag the text with
   # format - the format to use for the characters
   # begin - the start location
   # end - the end location (actually formats up to one char before)
   set markup_deleted 0

   set mark [MH_FindEndTag $widget $markup_key \
      $end "$search_start +${search_length}c"]
   set mark2(begin) [lindex $mark 0]
   if {$mark2(begin) != ""} {
      # End tag has been found
      set mark2(end) [lindex $mark 1]
      set mark2(text) [lindex $mark 2]
      $widget tag add $markup_key $begin "$mark2(end) +1c"
      if {$markup_format != ""} {
         eval $widget tag configure $markup_key $markup_format
      }
      # Perform tag deletion and bookkeeping
      $widget delete $mark2(begin) "$mark2(end) +1c"
      $widget delete $begin "$end +1c"
      incr markup_deleted [expr [string length $text] \
         + [string length $mark2(text)] +2]
   }

   # Return new markup deleted count
   return $markup_deleted
}

###############################################################
# The procedure for inserting a given text in place of a tag.
# May be used without an end tag.
proc MH_FormatReplace { widget markup_key markup_format begin end \
   text search_start search_length} {
   # Info_Dialog . "Key: $markup_key, Tag1: $begin ->$end"

   # widget - text widget containing the area to be formatted
   # tag - the string to tag the text with
   # format - the format to use for the characters
   # begin - the start location
   # end - the end location (actually formats up to one char before)
   set markup_deleted 0

   set mark [MH_FindEndTag $widget $markup_key \
      $end "$search_start +${search_length}c"]
   set mark2(begin) [lindex $mark 0]
   if {$mark2(begin) != ""} {
      # End tag has been found
      # Perform tag deletion and bookkeeping
      set mark2(end) [lindex $mark 1]
      set mark2(text) [lindex $mark 2]
      $widget delete $mark2(begin) "$mark2(end) +1c"
      incr markup_deleted [expr [string length $mark2(text)] +1]
   }
   $widget delete $begin "$end +1c"
   if {$markup_format != ""} {
      $widget insert $begin $markup_format
   }
   # Take account of insrted characters
   incr markup_deleted [expr [string length $text] +1 - \
      [string length $markup_format]]

   # Return new markup deleted count
   return $markup_deleted
}

###############################################################
# The procedure for formatting characters
proc MH_FormatImage { widget begin end text {image_path ""} } {
   # Info_Dialog . "Image: $text, Tag1: $begin ->$end"

   # widget - text widget containing the area to be formatted
   # begin - start of tagged text
   # end - end of first tag (-1 char)
   # text - the tag - contains the image name
   # image_path - optional list containing image path
   set markup_deleted 0

   set attribute_list [MH_GetAttributes $text]
   set attribute_count 0
   while {$attribute_count < [llength attribute_list]} {

      switch [lindex $attribute_list $attribute_count] {
         SRC {
            set image_name [lindex $attribute_list \
               [expr $attribute_count +1]]
            if {$image_name != ""} {         
               # Call image creation and insertion routine
               MH_InsertImage $widget $begin $image_name
            }
         }
         ALIGN {
            # Set image alignment
            switch [string toupper [lindex $attribute_list \
               [expr $attribute_count +1]]] {
               CENTER {
                  set invalid_code 0
                  set align CENTER
                  set align_format "-justify center"
               }
               LEFT {
                  set invalid_code 0
                  set align LEFT
                  set align_format "-justify left"
               }
               RIGHT {
                  set invalid_code 0
                  set align RIGHT
                  set align_format "-justify right"
               }
               default {set invalid_code 1}
            }
            if {$invalid_code == 0} {
               $widget tag add $align $begin $end
               eval $widget tag configure $align $align_format
            }
         }
         ALT {}
      }
      # Increment counter
      incr attribute_count 2
   }

   # Perform tag deletion and bookkeeping
   $widget delete "$begin +1c" "$end +2c"
   incr markup_deleted [expr [string length $text] +1]

   # Return new markup deleted count
   return $markup_deleted
}

###############################################################
# The procedure for formatting breaks and paragraphs
proc MH_FormatPara { widget markup_key markup_format begin end \
   text search_start search_length} {

   # widget - text widget containing the area to be formatted
   # tag - the string to tag the text with
   # format - the format to use for the characters
   # begin - the start location
   # end - the end location (actually formats up to one char before)
   set markup_deleted 0

   set mark [MH_FindEndTag $widget $markup_key \
      $end "$search_start +${search_length}c"]
   set mark2(begin) [lindex $mark 0]
   # Assume that there is an end tag until proved otherwise
   set no_end_tag 0
   if {$mark2(begin) == ""} {
      # No end tag was found - the beginning of the next paragraph
      # is the end of the current set of justification (if justificatiion
      # is on)
      set no_end_tag 1
      set mark [MH_FindNextStartTag $widget $markup_key \
         $end "$search_start +${search_length}c"]
      set mark2(begin) [lindex $mark 0]
      set mark2(end) [lindex $mark 0]
   } else {
      # End tag has been found
      set mark2(end) [lindex $mark 1]
      set mark2(text) [lindex $mark 2]
   }

   # Find any justification present 
   # Example: P ALIGN=CENTER
   set attribute_list [MH_GetAttributes $text]
   if {[llength $attribute_list] == 2} {
      if {$mark2(begin) != ""} {
   
         # Tag characters if justified
         switch [string toupper [lindex $attribute_list 1]] {
            CENTER {
               set invalid_code 0
               set align CENTER
               set align_format "-justify center"
            }
            LEFT {
               set invalid_code 0
               set align LEFT
               set align_format "-justify left"
            }
            RIGHT {
               set invalid_code 0
               set align RIGHT
               set align_format "-justify right"
            }
            default {set invalid_code 1}
         }
         if {$invalid_code == 0} {
            $widget tag add $align $begin "$mark2(end) +1c"
            eval $widget tag configure $align $align_format
         }
      }
   }

   # Perform tag deletion and bookkeeping if there was an end tag
   if {$no_end_tag == 0} {
      $widget delete $mark2(begin) "$mark2(end) +1c"
      incr markup_deleted [expr [string length $mark2(text)] +1]
   }

   # Note that inserted characters are not justfied
   $widget delete $begin "$end +1c"
   $widget insert $begin $markup_format

   # Take account of insrted characters
   incr markup_deleted [expr [string length $text] +1 - \
      [string length $markup_format]]

   # Return new markup deleted count
   return $markup_deleted
}


